@@ -219,6 +219,9 @@ typedef enum {
 	GTK_HTML_COMMAND_CELL_RSPAN_INC,
 	GTK_HTML_COMMAND_CELL_CSPAN_DEC,
 	GTK_HTML_COMMAND_CELL_RSPAN_DEC,
+
+	GTK_HTML_COMMAND_EDITABLE_ON,
+	GTK_HTML_COMMAND_EDITABLE_OFF,
 } GtkHTMLCommandType;
 
 typedef enum {
@@ -301,6 +304,8 @@ typedef enum {
 
 typedef enum {
 	GTK_HTML_BEGIN_KEEP_SCROLL = 1 << 0,
-	GTK_HTML_BEGIN_KEEP_IMAGES = 1 << 1
+	GTK_HTML_BEGIN_KEEP_IMAGES = 1 << 1,
+	GTK_HTML_BEGIN_BLOCK_UPDATES = 1 << 2,
+	GTK_HTML_BEGIN_BLOCK_IMAGES = 1 << 3
 } GtkHTMLBeginFlags;
 #endif
