@@ -316,17 +316,13 @@ html_image_real_calc_size (HTMLObject *o, HTMLPainter *painter, GList **changed_
 
 	pixel_size = html_painter_get_pixel_size (painter);
 
-	if (o->parent && HTML_IS_PLAIN_PAINTER (painter) && image->alt && *image->alt) {
-		GtkHTMLFontStyle style = GTK_HTML_FONT_STYLE_DEFAULT;
-		HTMLClueFlow *cf = html_object_get_flow (o);
-		gint lo = 0;
-
-		if (cf)
-			style = html_clueflow_get_default_font_style (cf);
-
+	if (o->parent && HTML_IS_CLUEFLOW (o->parent)
+	    && HTML_IS_PLAIN_PAINTER (painter) && image->alt && *image->alt) {
+		html_painter_set_font_style (painter, html_clueflow_get_default_font_style (HTML_CLUEFLOW (o->parent)));
+		html_painter_set_font_face (painter, NULL);
 		/* FIXME: cache items and glyphs? */
-		html_painter_calc_text_size (painter, image->alt, g_utf8_strlen (image->alt, -1), NULL, NULL, NULL, 0, &lo,
-					     style, NULL, &o->width, &o->ascent, &o->descent);
+		html_painter_calc_text_size (painter, image->alt, g_utf8_strlen (image->alt, -1),
+					     &o->width, &o->ascent, &o->descent);
 	} else {
 		width = html_image_get_actual_width (image, painter);
 		height = html_image_get_actual_height (image, painter);
@@ -356,7 +352,6 @@ draw_plain (HTMLObject *o, HTMLPainter *p, gint x, gint y, gint width, gint heig
 		return;
 
 	if (img->alt && *img->alt) {
-		HTMLClueFlow *cf = html_object_get_flow (o);
 
 		/* FIXME: cache items and glyphs? */
 		if (o->selected) {
@@ -371,11 +366,10 @@ draw_plain (HTMLObject *o, HTMLPainter *p, gint x, gint y, gint width, gint heig
 			html_painter_set_pen (p, &html_colorset_get_color_allocated (e->settings->color_set, p,
 										     HTMLTextColor)->color);
 		}
-		if (cf)
-			html_painter_set_font_style (p, html_clueflow_get_default_font_style (cf));
-
-  		html_painter_set_font_face (p, NULL);
-		html_painter_draw_text (p, o->x + tx, o->y + ty, img->alt, g_utf8_strlen (img->alt, -1), NULL, NULL, NULL, 0, 0);
+		
+		html_painter_set_font_style (p, html_clueflow_get_default_font_style (HTML_CLUEFLOW (o->parent)));
+		html_painter_set_font_face (p, NULL);
+		html_painter_draw_text (p, o->x + tx, o->y + ty, img->alt, g_utf8_strlen (img->alt, -1));
 	}
 }
 
