@@ -32,9 +32,12 @@
 #include "htmlcolor.h"
 #include "htmlcolorset.h"
 #include "htmlpainter.h"
+#include "htmlobject.h"
 #include "htmlprinter.h"
 #include "htmlplainpainter.h"
 #include "htmlgdkpainter.h"
+#include "htmlsettings.h"
+#include "gtkhtml.h"
 
 
 /* #define HTML_TEXT_SLAVE_DEBUG */
@@ -44,6 +47,7 @@ static HTMLObjectClass *parent_class = NULL;
 
 static GList * get_glyphs (HTMLTextSlave *slave, HTMLPainter *painter);
 static GList * get_glyphs_part (HTMLTextSlave *slave, HTMLPainter *painter, guint offset, guint len);
+static void    clear_glyphs (HTMLTextSlave *slave);
 
 char *
 html_text_slave_get_text (HTMLTextSlave *slave)
@@ -99,10 +103,9 @@ html_text_slave_get_start_byte_offset (HTMLTextSlave *slave)
 }
 
 static guint
-calc_width (HTMLTextSlave *slave, HTMLPainter *painter, gint *asc, gint *dsc)
+hts_calc_width (HTMLTextSlave *slave, HTMLPainter *painter, gint *asc, gint *dsc)
 {
 	HTMLText *text = slave->owner;
-	HTMLObject *next, *prev;
 	gint line_offset, tabs = 0, width = 0;
 
 	line_offset = html_text_slave_get_line_offset (slave, 0, painter);
@@ -110,7 +113,7 @@ calc_width (HTMLTextSlave *slave, HTMLPainter *painter, gint *asc, gint *dsc)
 		width += (html_text_text_line_length (html_text_slave_get_text (slave), &line_offset, slave->posLen, &tabs) - slave->posLen)*
 			html_painter_get_space_width (painter, html_text_get_font_style (text), text->face);
 
-	width += html_text_calc_part_width (text, painter, slave->posStart, slave->posLen, asc, dsc);
+	width += html_text_calc_part_width (text, painter, html_text_slave_get_text (slave), slave->posStart, slave->posLen, asc, dsc);
 
 	return width;
 }
@@ -125,57 +128,11 @@ glyphs_destroy (GList *glyphs)
 	g_list_free (glyphs);
 }
 
-static gint
-get_offset_for_bounded_width (HTMLTextSlave *slave, HTMLPainter *painter, gint *words, gint max_width)
-{
-	HTMLText *text = slave->owner;
-	gint upper = slave->posLen;
-	gint lower = 0;
-	gint off = 0;
-	gint len, width, asc, dsc;
-	gint line_offset = -1;
-	gchar *sep, *str;
-	char *buffer = html_text_slave_get_text (slave);
-
-	len = (lower + upper) / 2;
-	html_text_calc_text_size (slave->owner, painter, buffer, len, html_text_get_pango_info (text, painter),
-				  get_glyphs (slave, painter), html_text_slave_get_start_byte_offset (slave),
-				  &line_offset, html_text_get_font_style (text), text->face, &width, &asc, &dsc);
-	while (lower < upper) {
-		if (width > max_width)
-			upper = len - 1;
-		else
-			lower = len + 1;
-		len = (lower + upper) / 2;
-		line_offset = -1;
-		html_text_calc_text_size (slave->owner, painter, buffer, len, html_text_get_pango_info (text, painter), get_glyphs (slave, painter),
-					  html_text_slave_get_start_byte_offset (slave),
-					  &line_offset, html_text_get_font_style (text), text->face, &width, &asc, &dsc);
-	}
-
-	if (width > max_width && len > 1)
-		len --;
-
-	*words = 0;
-	str = sep = buffer;
-
-	while (off < len && *sep) {
-		if (*sep == ' ') 
-			(*words) ++;
-
-		sep = g_utf8_next_char (sep);
-		off++;
-	}
-
-	return len;
-}
-
 static gboolean
-calc_size (HTMLObject *self, HTMLPainter *painter, GList **changed_objs)
+html_text_slave_real_calc_size (HTMLObject *self, HTMLPainter *painter, GList **changed_objs)
 {
 	HTMLText *owner;
 	HTMLTextSlave *slave;
-	HTMLObject *next;
 	GtkHTMLFontStyle font_style;
 	gint new_ascent, new_descent, new_width;
 	gboolean changed;
@@ -184,7 +141,7 @@ calc_size (HTMLObject *self, HTMLPainter *painter, GList **changed_objs)
 	owner = HTML_TEXT (slave->owner);
 	font_style = html_text_get_font_style (owner);
 
-	new_width = MAX (1, calc_width (slave, painter, &new_ascent, &new_descent));
+	new_width = MAX (1, hts_calc_width (slave, painter, &new_ascent, &new_descent));
 
 	/* handle sub & super script */
 	if (font_style & GTK_HTML_FONT_STYLE_SUBSCRIPT || font_style & GTK_HTML_FONT_STYLE_SUPERSCRIPT) {
@@ -312,28 +269,32 @@ html_text_slave_get_nb_width (HTMLTextSlave *slave, HTMLPainter *painter, gboole
 	return html_object_calc_min_width (HTML_OBJECT (slave), painter);
 }
 
+/*
+ * offset: current position - char offset
+ * s: current position - string pointer
+ * ii: current position - index of item
+ * io: current position - offset within item
+ * line_offset: 
+ * w: width for current position
+ * lwl: last-whitespacing-length (in characters)
+ * lbw: last-break-width
+ * lbo: last-break-offset (into text)
+ * lbsp: last-break-string-pointer
+ *
+ * Checks to see if breaking the text at the given position would result in it fitting
+ * into the remaining width. If so, updates the last-break information (lwl,lbw,lbo,lbsp).
+ * We'll actually break the item at the last break position that still fits.
+ */
 static gboolean
 update_lb (HTMLTextSlave *slave, HTMLPainter *painter, gint widthLeft, gint offset, gchar *s, gint ii, gint io, gint line_offset,
-	   gint *w, gint *ltw, gint *lwl, gint *lbw, gint *lbo, gchar **lbsp, gboolean *force_fit)
+	   gint *w, gint *lwl, gint *lbw, gint *lbo, gchar **lbsp, gboolean *force_fit)
 {
 	gint new_ltw, new_lwl, aw;
 
 	new_ltw = html_text_tail_white_space (slave->owner, painter, offset, ii, io, &new_lwl, line_offset, s);
-	if (HTML_IS_GDK_PAINTER (painter) || HTML_IS_PLAIN_PAINTER (painter)) {
-		aw = *w - new_ltw;
-	} else {
-		gint lo = html_text_get_line_offset (slave->owner, painter, *lbo - *lwl);
-				/* printf ("s: %s l: %d\n", html_text_get_text (slave->owner, lbo - lwl), offset - new_lwl - lbo + lwl); */
-		html_text_calc_text_size (slave->owner, painter, html_text_get_text (slave->owner, *lbo - *lwl),
-					  offset - new_lwl - *lbo + *lwl, NULL, NULL, 0, &lo,
-					  html_text_get_font_style (slave->owner), slave->owner->face,
-					  &aw, NULL, NULL);
-		*w += aw;
-		aw = *w;
-		new_ltw = 0;
-	}
+	aw = *w - new_ltw;
+	
 	if (aw <= widthLeft || *force_fit) {
-		*ltw = new_ltw;
 		*lwl = new_lwl;
 		*lbw = aw;
 		*lbo = offset;
@@ -352,7 +313,7 @@ hts_fit_line (HTMLObject *o, HTMLPainter *painter,
 	      gboolean lineBegin, gboolean firstRun, gboolean next_to_floating, gint widthLeft)
 {
 	HTMLTextSlave *slave = HTML_TEXT_SLAVE (o);
-	gint lbw, w, lbo, ltw, lwl, offset;
+	gint lbw, w, lbo, lwl, offset;
 	gint ii, io, line_offset;
 	gchar *s, *lbsp;
 	HTMLFitType rv = HTML_FIT_NONE;
@@ -362,59 +323,43 @@ hts_fit_line (HTMLObject *o, HTMLPainter *painter,
 	if (rv == HTML_FIT_COMPLETE)
 		return rv;
 
-	lbw = ltw = lwl = w = 0;
+	widthLeft = html_painter_engine_to_pango (painter, widthLeft);
+
+	lbw = lwl = w = 0;
 	offset = lbo = slave->posStart;
 	ii = html_text_get_item_index (slave->owner, painter, offset, &io);
 
 	line_offset = html_text_get_line_offset (slave->owner, painter, offset);
-	lbsp = s = html_text_get_text (slave->owner, offset);
+	lbsp = s = html_text_slave_get_text (slave);
 
 	while ((force_fit || widthLeft > lbw) && offset < slave->posStart + slave->posLen) {
-		if (offset > slave->posStart && pi->entries [ii].attrs [io].is_line_break)
-			if (update_lb (slave, painter, widthLeft, offset, s, ii, io, line_offset, &w, &ltw, &lwl, &lbw, &lbo, &lbsp, &force_fit))
+		if (offset > slave->posStart && offset > lbo && html_text_is_line_break (pi->attrs [offset]))
+			if (update_lb (slave, painter, widthLeft, offset, s, ii, io, line_offset, &w, &lwl, &lbw, &lbo, &lbsp, &force_fit))
 				break;
 
 		if (*s == '\t') {
 			gint skip = 8 - (line_offset % 8);
-			if (HTML_IS_GDK_PAINTER (painter) || HTML_IS_PLAIN_PAINTER (painter))
-				w += skip*PANGO_PIXELS (pi->entries [ii].widths [io]);
+			w += skip*pi->entries [ii].widths [io];
 			line_offset += skip;
 		} else {
-			if (HTML_IS_GDK_PAINTER (painter) || HTML_IS_PLAIN_PAINTER (painter))
-				w += PANGO_PIXELS (pi->entries [ii].widths [io]);
+			w += pi->entries [ii].widths [io];
 			line_offset ++;
 		}
 
 		s = g_utf8_next_char (s);
 		offset ++;
 
-		if (offset > slave->posStart && ii != pi->n - 1 && io == pi->entries [ii].item->num_chars - 1)
-			if (update_lb (slave, painter, widthLeft, offset, s, ii, io, line_offset, &w, &ltw, &lwl, &lbw, &lbo, &lbsp, &force_fit))
-				break;
-
 		html_text_pi_forward (pi, &ii, &io);
 	}
 
-	if (!HTML_IS_GDK_PAINTER (painter) && !HTML_IS_PLAIN_PAINTER (painter)) {
-		gint aw;
-		gint lo = html_text_get_line_offset (slave->owner, painter, lbo - lwl);
-
-		/* printf ("s: %s l: %d\n", html_text_get_text (slave->owner, lbo - lwl), offset - lbo + lwl); */
-		html_text_calc_text_size (slave->owner, painter, html_text_get_text (slave->owner, lbo - lwl),
-					  offset - lbo + lwl, NULL, NULL, 0, &lo,
-					  html_text_get_font_style (slave->owner), slave->owner->face,
-					  &aw, NULL, NULL);
-		w += aw;
-	}
-
 	if (offset == slave->posStart + slave->posLen && (widthLeft >= w || force_fit)) {
 		rv = HTML_FIT_COMPLETE;
 		if (slave->posLen)
-			o->width = w;
+			o->width = html_painter_pango_to_engine (painter, w);
 	} else if (lbo > slave->posStart) {
 		split (slave, lbo - slave->posStart, lbsp);
 		rv = HTML_FIT_PARTIAL;
-		o->width = lbw;
+		o->width = html_painter_pango_to_engine (painter, lbw);
 		slave->posLen -= lwl;
 	}
 
@@ -465,6 +410,12 @@ draw_spell_errors (HTMLTextSlave *slave, HTMLPainter *p, gint tx, gint ty)
 	gint last_off = 0;
 	gint line_offset = html_text_slave_get_line_offset (slave, 0, p);
 	gchar *text = html_text_slave_get_text (slave);
+	HTMLEngine *e;
+
+	if (p->widget && GTK_IS_HTML (p->widget))
+		e = GTK_HTML (p->widget)->engine;
+	else
+		return;
 
 	while (cur) {
 
@@ -475,25 +426,23 @@ draw_spell_errors (HTMLTextSlave *slave, HTMLPainter *p, gint tx, gint ty)
 			GList *glyphs;
 			guint off = ma - slave->posStart;
 			guint len = mi - ma;
-			gint lo, width, asc, dsc;
+			gint width, asc, dsc;
 
-			html_painter_set_pen (p, &html_colorset_get_color_allocated (p, HTMLSpellErrorColor)->color);
+			html_painter_set_pen (p, &html_colorset_get_color_allocated (e->settings->color_set,
+										     p, HTMLSpellErrorColor)->color);
 			/* printf ("spell error: %s\n", html_text_get_text (slave->owner, off)); */
-			lo = line_offset;
 			
 			glyphs = get_glyphs_part (slave, p, last_off, off - last_off);
-			html_text_calc_text_size (slave->owner, p, text,
-						  off - last_off, html_text_get_pango_info (slave->owner, p), glyphs, text - slave->owner->text,
-						  &line_offset,
-						  p->font_style,
-						  p->font_face, &width, &asc, &dsc);
+			html_text_calc_text_size (slave->owner, p, text - slave->owner->text,
+						  off - last_off, html_text_get_pango_info (slave->owner, p), glyphs,
+						  &line_offset, &width, &asc, &dsc);
 			glyphs_destroy (glyphs);
 			x_off += width;
 			text = g_utf8_offset_to_pointer (text, off - last_off);
 			glyphs = get_glyphs_part (slave, p, off, len);
 			x_off += html_painter_draw_spell_error (p, obj->x + tx + x_off,
 								obj->y + ty + get_ys (slave->owner, p),
-								text, len, html_text_get_pango_info (slave->owner, p), glyphs, text - slave->owner->text);
+								html_text_get_pango_info (slave->owner, p), glyphs);
 			glyphs_destroy (glyphs);
 			last_off = off + len;
 			if (line_offset != -1)
@@ -506,29 +455,6 @@ draw_spell_errors (HTMLTextSlave *slave, HTMLPainter *p, gint tx, gint ty)
 	}
 }
 
-static gint
-get_pango_info_offset (HTMLTextSlave *slave, HTMLPainter *painter)
-{
-	gint pi_offset = 0;
-
-	if (!slave->pi) {
-		PangoItem *item;
-		gint start_offset = html_text_slave_get_text (slave) - slave->owner->text;
-
-		slave->pi = html_text_get_pango_info (slave->owner, painter);
-		if (slave->pi) {
-			item = slave->pi->entries [pi_offset].item;
-
-			while (start_offset >= item->offset + item->length) {
-				pi_offset ++;
-				item = slave->pi->entries [pi_offset].item;
-			}
-		}
-	}
-
-	return pi_offset;
-}
-
 static inline GList *
 get_glyphs_base_text (GList *glyphs, PangoItem *item, gint ii, const gchar *text, gint bytes)
 {
@@ -601,6 +527,8 @@ static GList *
 get_glyphs (HTMLTextSlave *slave, HTMLPainter *painter)
 {
 	if (!slave->glyphs || (HTML_OBJECT (slave)->change & HTML_CHANGE_RECALC_PI)) {
+		clear_glyphs (slave);
+
 		HTML_OBJECT (slave)->change &= ~HTML_CHANGE_RECALC_PI;
 		slave->glyphs = get_glyphs_part (slave, painter, 0, slave->posLen);
 	}
@@ -625,28 +553,17 @@ draw_normal (HTMLTextSlave *self,
 
 	str = html_text_slave_get_text (self);
 	if (*str) {
-		GList *glyphs, *items;
-		PangoAttrList *attrs = NULL;
+		GList *glyphs;
 
 		if (self->posStart > 0)
 			glyphs = get_glyphs_part (self, p, 0, self->posLen);
 		else
 			glyphs = get_glyphs (self, p);
 
-		if (HTML_IS_PRINTER (p)) {
-			gchar *text = html_text_slave_get_text (self);
-			gint start_index, end_index;
-
-			start_index = html_text_slave_get_text (self) - self->owner->text;
-			end_index = g_utf8_offset_to_pointer (text, self->posLen) - text;
-			attrs = html_text_get_attr_list (self->owner, start_index, end_index);
-		}
-
-		html_painter_draw_text (p, obj->x + tx, obj->y + ty + get_ys (text, p),
-					str, self->posLen, html_text_get_pango_info (text, p), attrs, glyphs,
-					str - self->owner->text, html_text_slave_get_line_offset (self, 0, p));
-		if (attrs)
-			pango_attr_list_unref (attrs);
+		html_painter_draw_entries (p, obj->x + tx, obj->y + ty + get_ys (text, p),
+					   str, self->posLen,
+					   html_text_get_pango_info (text, p), glyphs,
+					   html_text_slave_get_line_offset (self, 0, p));
 
 		if (self->posStart > 0)
 			glyphs_destroy (glyphs);
@@ -659,6 +576,12 @@ draw_focus  (HTMLPainter *painter, GdkRectangle *box)
 	HTMLGdkPainter *p;
 	GdkGCValues values;
 	gchar dash [2];
+	HTMLEngine *e;
+
+	if (painter->widget && GTK_IS_HTML (painter->widget))
+		e = GTK_HTML (painter->widget)->engine;
+	else
+		return;
 
 	if (HTML_IS_PRINTER (painter))
 		return;
@@ -666,7 +589,8 @@ draw_focus  (HTMLPainter *painter, GdkRectangle *box)
 	p = HTML_GDK_PAINTER (painter);
 	/* printf ("draw_text_focus\n"); */
 
-	gdk_gc_set_foreground (p->gc, &html_colorset_get_color_allocated (painter, HTMLTextColor)->color);
+	gdk_gc_set_foreground (p->gc, &html_colorset_get_color_allocated (e->settings->color_set,
+									  painter, HTMLTextColor)->color);
 	gdk_gc_get_values (p->gc, &values);
 
 	dash [0] = 1;
@@ -715,11 +639,11 @@ draw (HTMLObject *o,
 			gint bw = 0;
 			html_object_get_bounds (o, &rect);
 			if (textslave->posStart < link->start_offset)
-				bw = html_text_calc_part_width (owner, p, textslave->posStart, link->start_offset - textslave->posStart, NULL, NULL);
+				bw = html_text_calc_part_width (owner, p, html_text_slave_get_text (textslave), textslave->posStart, link->start_offset - textslave->posStart, NULL, NULL);
 			rect.x += tx + bw;
 			rect.width -= bw;
 			if (textslave->posStart + textslave->posLen > link->end_offset)
-				rect.width -= html_text_calc_part_width (owner, p, link->end_offset,  textslave->posStart + textslave->posLen - link->end_offset, NULL, NULL);
+				rect.width -= html_text_calc_part_width (owner, p, owner->text + link->end_index, link->end_offset,  textslave->posStart + textslave->posLen - link->end_offset, NULL, NULL);
 			rect.y += ty;
 			draw_focus (p, &rect);
 		}
@@ -753,14 +677,12 @@ static gint
 calc_offset (HTMLTextSlave *slave, HTMLPainter *painter, gint x, gint y)
 {
 	HTMLText *owner;
-	GtkHTMLFontStyle font_style;
 	gint line_offset;
 	guint width, prev_width;
 	gchar *text;
 	guint upper;
 	guint len;
 	guint lower;
-	GList *items;
 	GList *glyphs;
 	gint lo;
 	gint asc, dsc;
@@ -776,7 +698,6 @@ calc_offset (HTMLTextSlave *slave, HTMLPainter *painter, gint x, gint y)
 	text = html_text_slave_get_text (slave);
 	line_offset = html_text_slave_get_line_offset (slave, 0, painter);	
 	owner = HTML_TEXT (slave->owner);
-	font_style = html_text_get_font_style (owner);
 
 	while (upper - lower > 1) {
 		lo = line_offset;
@@ -791,9 +712,8 @@ calc_offset (HTMLTextSlave *slave, HTMLPainter *painter, gint x, gint y)
 
 		if (len) {
 			glyphs = get_glyphs_part (slave, painter, 0, len);
-			html_text_calc_text_size (slave->owner, painter, text, len, html_text_get_pango_info (owner, painter), glyphs,
-						  html_text_slave_get_start_byte_offset (slave), 
-						  &lo, font_style, owner->face, &width, &asc, &dsc);
+			html_text_calc_text_size (slave->owner, painter, text - slave->owner->text, len, html_text_get_pango_info (owner, painter), glyphs,
+						  &lo, &width, &asc, &dsc);
 			glyphs_destroy (glyphs);
 		} else {
 			width = 0;
@@ -849,14 +769,20 @@ check_point (HTMLObject *self,
 }
 
 static void
-destroy (HTMLObject *obj)
+clear_glyphs (HTMLTextSlave *slave)
 {
-	HTMLTextSlave *slave = HTML_TEXT_SLAVE (obj);
-
 	if (slave->glyphs) {
 		glyphs_destroy (slave->glyphs);
 		slave->glyphs = NULL;
 	}
+}
+
+static void
+destroy (HTMLObject *obj)
+{
+	HTMLTextSlave *slave = HTML_TEXT_SLAVE (obj);
+
+	clear_glyphs (slave);
 
 	HTML_OBJECT_CLASS (parent_class)->destroy (obj);
 }
@@ -882,7 +808,7 @@ html_text_slave_class_init (HTMLTextSlaveClass *klass,
 	object_class->copy = copy;
 	object_class->destroy = destroy;
 	object_class->draw = draw;
-	object_class->calc_size = calc_size;
+	object_class->calc_size = html_text_slave_real_calc_size;
 	object_class->fit_line = hts_fit_line;
 	object_class->calc_min_width = calc_min_width;
 	object_class->calc_preferred_width = calc_preferred_width;
